#! /usr/local/bin/perl

use strict;
use utf8;
use Encode;

# 環境変数LANGが*.UTF-8のときはUTF-8、それ以外はCP932でメッセージ表示します
my $os_enc = 'cp932';
if ($ENV{LANG} =~ m/.+\.UTF-8$/) {
	$os_enc = 'UTF-8';
}

binmode STDOUT, ":encoding($os_enc)";
binmode STDERR, ":encoding($os_enc)";

sub DEC { Encode::decode($os_enc, shift) }

my $text_file = 'object_names.txt';

my $in_dir = '';
my $npc_file                = 'nsc.csv';
my $inventory_object_file   = 'inventarobjekt.csv';
my $object_interaction_file = 'raumobjektinteraktion.csv';
my $topic_file              = 'topic.csv';

my $opt_ignore_space_only = 1;
my $opt_check_same_text = 1;

sub usage()
{
	print "使用方法: get_object_names.pl [オプション]\n";
	print "機    能: nsc.csv, inventarobjekt.csv, raumobjektinteraktion.csv, topic.csvから\n";
	print "          オブジェクト名を抽出します\n";
	print "オプション:\n";
	print "    -o|--output ファイル名  指定したファイルに抽出したオブジェクト名を保存します\n";
	print "                            (デフォルト: object_names.txt)\n";
	print "    -i|--indir  ディレクトリ名\n";
	print "                            指定したディレクトリからスクリプトを読み込みます\n";
	print "\n";
	print "       --ignore-space-only  空白文字だけのテキストは出力しません\n";
	print "                            (デフォルト)\n";
	print "       --no-ignore-space-only\n";
	print "                            空白文字だけのテキストも出力します\n";
	print "\n";
	print "       --check-same-text    同じ原文テキストがあるとき翻訳に0と出力する\n";
	print "                            (デフォルト)\n";
	print "       --no-check-same-text 同じ原文テキストがあるとき翻訳に0と出力しない\n";
	print "\n";
	print "    -?|--help               使用方法を表示します\n";
}

while (my $var = Encode::decode($os_enc, shift @ARGV)) {
	if ($var =~ /^-/) {
		if ($var eq '-?' or $var eq '--help') {
			usage();
			exit(1);
		} elsif ($var eq '-o' or $var eq '--output') {
			if ($#ARGV < 0) {
				printf("ERROR: オプション%sの後にファイル名がありません\n", $var);
				exit(1);
			}
			$text_file = Encode::decode($os_enc, shift @ARGV);
		} elsif ($var eq '-i' or $var eq '--indir') {
			if ($#ARGV < 0) {
				printf("ERROR: オプション%sの後にディレクトリ名がありません\n", $var);
				exit(1);
			}
			$in_dir = Encode::decode($os_enc, shift @ARGV);
		} elsif ($var eq '--ignore-space-only') {
			$opt_ignore_space_only = 1;
		} elsif ($var eq '--no-ignore-space-only') {
			$opt_ignore_space_only = 0;
		} elsif ($var eq '--check-same-text') {
			$opt_check_same_text = 1;
		} elsif ($var eq '--no-check-same-text') {
			$opt_check_same_text = 0;
		} else {
			printf("ERROR: 不正なオプションです: %s\n", $var);
			exit(1);
		}
	} else {
		printf("ERROR: 不要なコマンドライン引数です: %s\n", $var);
		exit(1);
	}
}

if ($in_dir ne '') {
	$npc_file                = "$in_dir/$npc_file";
	$inventory_object_file   = "$in_dir/$inventory_object_file";
	$object_interaction_file = "$in_dir/$object_interaction_file";
	$topic_file              = "$in_dir/$topic_file";
}

my %object_names;
my $num_npc_names = 0;
my $num_inventory_object_names = 0;
my $num_object_interaction_names = 0;
my $num_topic_names = 0;

sub add_object_name($$$$)
{
	my ($type, $line, $id, $text) = @_;
	my $key = "$type\t$id";

	$text =~ s/[ \t]+$//;

	if (exists $object_names{$key}) {
		my ($prev_line, $prev_text) = split("\t", $object_names{$key}, 2);
		printf("WARNING: 同じ番号の名前が存在します: %d行目, %d行目: %s %s\n", $prev_line, $line, $id);
	} elsif ($opt_ignore_space_only != 0 and $text =~ m/^[ \t]*$/) {
		# 空白文字だけのテキストは登録しない
	} else {
		if ($type eq 'npc') {
			$num_npc_names++;
		} elsif ($type eq 'inventory') {
			$num_inventory_object_names++;
		} elsif ($type eq 'object') {
			$num_object_interaction_names++;
		} elsif ($type eq 'topic') {
			$num_topic_names++;
		} else {
			die "BUG: unknown type: $type\n";
		}

		$object_names{$key} = join("\t", $line, $text);
	}
}

print "NPCファイル読み込み中...\n";
open(my $NPC, '<:unix:encoding(UTF-8)', $npc_file) or die "ERROR: ファイルを開けません: $npc_file: " . DEC($!) . "\n";
my $line = 0;
while (<$NPC>) {
	$line++;
	if (m#^([0-9]+);[0-9]+;[0-9]+;([^;]+);[^;]+;[0-9]+(?:\.[0-9]+)?;[0-9]+(?:\.[0-9]+)?;[0-9]+;[0-9]+[ \t]*[\r\n]+$#) {
		add_object_name('npc', $line, $1, $2);
	}
}
close($NPC);

print "持ち物ファイル読み込み中...\n";
open(my $INVENTORY, '<:unix:encoding(UTF-8)', $inventory_object_file) or die "ERROR: ファイルを開けません: $inventory_object_file: " . DEC($!) . "\n";
$line = 0;
while (<$INVENTORY>) {
	$line++;
	if (m#^([0-9]+);[0-9]+;([^;]+);(?:items|topics)/[^;/]+;[0-9]+;[^;]+;[0-9]+;[0-9]+;[0-9]+[ \t]*[\r\n]+$#) {
		add_object_name('inventory', $line, $1, $2);
	}
}
close($INVENTORY);

print "オブジェクトファイル読み込み中...\n";
open(my $OBJECT, '<:unix:encoding(UTF-8)', $object_interaction_file) or die "ERROR: ファイルを開けません: $object_interaction_file: " . DEC($!) . "\n";
$line = 0;
while (<$OBJECT>) {
	$line++;
	if (m#^([0-9]+);[0-9]+;([^;]+);[0-9]+;[0-9]+;[^;]+;[^;]+;[0-9]+;[0-9]+;[0-9]+;[0-9]+[ \t]*[\r\n]+$#) {
		add_object_name('object', $line, $1, $2);
	}
}
close($OBJECT);

print "話題ファイル読み込み中...\n";
open(my $TOPIC, '<:unix:encoding(UTF-8)', $topic_file) or die "ERROR: ファイルを開けません: $topic_file: " . DEC($!) . "\n";
$line = 0;
while (<$TOPIC>) {
	$line++;
	if (m#^([0-9]+);[0-9]+;([^;]+);topics/[^;/]+;[0-9]+;[0-9]+;[0-9]+[ \t]*[\r\n]+$#) {
		add_object_name('topic', $line, $1, $2);
	}
}
close($TOPIC);

my %text_hash;

print "テキスト出力中...\n";
open(my $OUT, '>:unix:encoding(UTF-8)', $text_file) or die "ERROR: ファイルを開けません: $text_file: " . DEC($!) . "\n";
print $OUT "# 種別\tID\t原文\t翻訳\t翻訳\r\n";

sub compare_key($$)
{
	my ($key1, $key2) = @_;
	my ($type1, $id1) = split("\t", $key1);
	my ($type2, $id2) = split("\t", $key2);

	if ($type1 eq $type2) {
		return $id1 cmp $id2;
	} elsif ($type1 eq 'npc') {
		return -1;
	} elsif ($type1 eq 'inventory') {
		if ($type2 eq 'npc') {
			return 1;
		}
		return -1;
	} elsif ($type1 eq 'object') {
		if ($type2 eq 'topic') {
			return -1;
		}
		return 1;
	} elsif ($type1 eq 'topic') {
		return 1;
	}

	# ここに来たらバグってる
	return $key1 cmp $key2;
}

foreach my $key ( sort compare_key keys %object_names ) {
	my ($type, $id) = split("\t", $key, 2);
	my ($line, $text) = split("\t", $object_names{$key}, 2);

	my $ja_text = '';
	if ($opt_check_same_text != 0) {
		if (exists $text_hash{$text}) {
			$ja_text = '0';
		} else {
			$text_hash{$text} = 1;
		}
	}

	printf($OUT "%s\t%s\t%s\t%s\t\r\n", $type, $id, $text, $ja_text);
}

close($OUT);

print "完了\n\n";

printf("NPC       : %d件\n", $num_npc_names);
printf("持ち物    : %d件\n", $num_inventory_object_names);
printf("オジェクト: %d件\n", $num_object_interaction_names);
printf("話題      : %d件\n", $num_topic_names);

