#! /usr/local/bin/perl

use strict;
use utf8;
use Encode;

# 環境変数LANGが*.UTF-8のときはUTF-8、それ以外はCP932でメッセージ表示します
my $os_enc = 'cp932';
if ($ENV{LANG} =~ m/.+\.UTF-8$/) {
	$os_enc = 'UTF-8';
}

binmode STDOUT, ":encoding($os_enc)";
binmode STDERR, ":encoding($os_enc)";

sub DEC { Encode::decode($os_enc, shift) }

my $text_file = 'script_text.txt';
my $in_dir = '';
my $script_file = 'skript.csv';
my $choice_file = 'choiceliste.csv';

my $script_encoding = 'iso-8859-1';

my $opt_ignore_space_only = 1;
my $opt_check_same_text = 1;

sub usage()
{
	print "使用方法: get_script_text.pl [オプション]\n";
	print "機    能: skript.csv, choiceliste.csvからテキストを抽出します\n";
	print "オプション:\n";
	print "       --latin1             skript.csvをLatin-1(iso-8859-1)で読み込みます\n";
	print "                            英語のファイルを使用する場合に指定してください\n";
	print "\n";
	print "    -o|--output ファイル名  指定したファイルに抽出したテキストを保存します\n";
	print "                            (デフォルト: script_text.txt)\n";
	print "    -i|--indir  ディレクトリ名\n";
	print "                            指定したディレクトリからスクリプトを読み込みます\n";
	print "\n";
	print "       --ignore-space-only  空白文字だけのテキストは出力しません\n";
	print "                            (デフォルト)\n";
	print "       --no-ignore-space-only\n";
	print "                            空白文字だけのテキストも出力します\n";
	print "\n";
	print "       --check-same-text    同じ原文テキストがあるとき翻訳に0と出力する\n";
	print "                            (デフォルト)\n";
	print "       --no-check-same-text 同じ原文テキストがあるとき翻訳に0と出力しない\n";
	print "\n";
	print "    -?|--help               使用方法を表示します\n";
}

while (my $var = Encode::decode($os_enc, shift @ARGV)) {
	if ($var =~ /^-/) {
		if ($var eq '-?' or $var eq '--help') {
			usage();
			exit(1);
		} elsif ($var eq '--latin1') {
			$script_encoding = 'iso-8859-1';
		} elsif ($var eq '-o' or $var eq '--output') {
			if ($#ARGV < 0) {
				printf("ERROR: オプション%sの後にファイル名がありません\n", $var);
				exit(1);
			}
			$text_file = Encode::decode($os_enc, shift @ARGV);
		} elsif ($var eq '-i' or $var eq '--indir') {
			if ($#ARGV < 0) {
				printf("ERROR: オプション%sの後にディレクトリ名がありません\n", $var);
				exit(1);
			}
			$in_dir = Encode::decode($os_enc, shift @ARGV);
		} elsif ($var eq '--ignore-space-only') {
			$opt_ignore_space_only = 1;
		} elsif ($var eq '--no-ignore-space-only') {
			$opt_ignore_space_only = 0;
		} elsif ($var eq '--check-same-text') {
			$opt_check_same_text = 1;
		} elsif ($var eq '--no-check-same-text') {
			$opt_check_same_text = 0;
		} else {
			printf("ERROR: 不正なオプションです: %s\n", $var);
			exit(1);
		}
	} else {
		printf("ERROR: 不要なコマンドライン引数です: %s\n", $var);
		exit(1);
	}
}

if ($in_dir ne '') {
	$script_file = "$in_dir/$script_file";
	$choice_file = "$in_dir/$choice_file";
}

my %script_text;
my $num_script_text = 0;
my $num_choice_text = 0;

sub add_script_text($$$$$$)
{
	my ($type, $line, $id1, $id2, $func, $text) = @_;
	my $key = "$type\t$id1\t$id2";

	if (exists $script_text{$key}) {
		my ($prev_line, $prev_type, $prev_text) = split("\t", $script_text{$key}, 3);
		printf("WARNING: 同じ番号のテキストが存在します: %d行目, %d行目: %s %s\n", $prev_line, $line, $id1, $id2);
	} elsif ($opt_ignore_space_only != 0 and $text =~ m/^[ \t]*$/) {
		# 空白文字だけのテキストは登録しない
	} else {
		$script_text{$key} = join("\t", $line, $func, $text);
		if ($type eq 'script') {
			$num_script_text++;
		} elsif ($type eq 'choice') {
			$num_choice_text++;
		} else {
			die "BUG: unknown type: $type\n";
		}
	}
}

print "スクリプトファイル読み込み中...\n";
open(my $SCRIPT, "<:unix:encoding($script_encoding)", $script_file) or die "ERROR: ファイルを開けません: $script_file: " . DEC($!) . "\n";
my $line = 0;
while (<$SCRIPT>) {
	$line++;
	if (m/^([0-9]+);([0-9]+);(if(?:item)?active\([0-9]+\),)?(say|think)\((.+)\);+([^;]*[\r\n]+)$/) {
		add_script_text('script', $line, $1, $2, $4, $5);
	} elsif (m/^([0-9]+);([0-9]+);(if(?:item)?active\([0-9]+\),)?(saynscp?)\(([0-9]+),(.+)\);+([^;]*[\r\n]+)$/) {
		add_script_text('script', $line, $1, $2, $4, $6);
	} elsif (m/^([0-9]+);([0-9]+);(if(?:item)?active\([0-9]+\),)?(saysound)\(([0-9]+,[0-9]+,|char,)?(.+)\);+([^;]*[\r\n]+)$/) {
		add_script_text('script', $line, $1, $2, $4, $6);
	} elsif (m/^([0-9]+);([0-9]+);(if(?:item)?active\([0-9]+\),)?(saysoundfile)\(((?:[0-9]+,[0-9]+,[^,]+|char(?:,[^,]+)?)),(.+)\);+([^;]*[\r\n]+)$/) {
		add_script_text('script', $line, $1, $2, $4, $6);
	}
}
close($SCRIPT);

print "選択肢ファイル読み込み中...\n";
open(my $CHOICE, '<:unix:encoding(UTF-8)', $choice_file) or die "ERROR: ファイルを開けません: $choice_file: " . DEC($!) . "\n";
$line = 0;
while (<$CHOICE>) {
	$line++;
	if (m/^([0-9]+);([0-9]+);([^;]+);(.+);([0-9]+)([ \t]*[\r\n]+)$/) {
		add_script_text('choice', $line, $1, $2, $5, $4);
	}
}
close($CHOICE);

my %text_hash;

print "テキスト出力中...\n";
open(my $OUT, '>:unix:encoding(UTF-8)', $text_file) or die "ERROR: ファイルを開けません: $text_file: " . DEC($!) . "\n";
print $OUT "# 種別\tID\tID2\t関数/ID\t原文\t翻訳\t翻訳\r\n";

sub compare_key($$)
{
	my ($key1, $key2) = @_;
	my ($type1, $k1id1, $k1id2) = split("\t", $key1);
	my ($type2, $k2id1, $k2id2) = split("\t", $key2);

	if ($type1 eq $type2) {
		if ($type1 eq 'script') {
			if ($k1id1 ne $k2id1) {
				return $k1id1 cmp $k2id1;
			}
			return $k1id2 <=> $k2id2;
		} elsif ($type1 eq 'choice') {
			return $key1 cmp $key2;
		}
	} elsif ($type1 eq 'script') {
		return -1;
	} elsif ($type1 eq 'choice') {
		return 1;
	}

	# ここに来たらバグってる
	return $key1 cmp $key2;
}

foreach my $key ( sort compare_key keys %script_text ) {
	my ($type, $id1, $id2) = split("\t", $key, 3);
	my ($line, $func, $text) = split("\t", $script_text{$key}, 3);

	my $ja_text = '';
	if ($opt_check_same_text != 0) {
		# 重複判定は前後の空白文字を除去して行う
		my $text2 = $text;
		$text2 =~ s/^[ \t]+//;
		$text2 =~ s/[ \t]+$//;
		if (exists $text_hash{$text2}) {
			$ja_text = '0';
		} else {
			$text_hash{$text2} = 1;
		}
	}

	printf($OUT "%s\t%s\t%s\t%s\t%s\t%s\t\r\n", $type, $id1, $id2, $func, $text, $ja_text);
}

print "完了\n\n";

printf("スクリプト: %d件\n", $num_script_text);
printf("選択肢    : %d件\n", $num_choice_text);

