#! /usr/local/bin/perl

use strict;
use utf8;
use Encode;

# 環境変数LANGが*.UTF-8のときはUTF-8、それ以外はCP932でメッセージ表示します
my $os_enc = 'cp932';
if ($ENV{LANG} =~ m/.+\.UTF-8$/) {
	$os_enc = 'UTF-8';
}

binmode STDOUT, ":encoding($os_enc)";
binmode STDERR, ":encoding($os_enc)";

sub DEC { Encode::decode($os_enc, shift) }

my $text_file = 'object_names.txt';

my $in_dir = '';
my $npc_file                = 'nsc.csv';
my $inventory_object_file   = 'inventarobjekt.csv';
my $object_interaction_file = 'raumobjektinteraktion.csv';
my $topic_file              = 'topic.csv';

my $out_dir = 'SCRIPT_NEW';
my $out_npc_file                = $npc_file;
my $out_inventory_object_file   = $inventory_object_file;
my $out_object_interaction_file = $object_interaction_file;
my $out_topic_file              = $topic_file;

sub usage()
{
	print "使用方法: replace_object_names.pl [オプション]\n";
	print "機    能: nsc.csv, inventarobjekt.csv, raumobjektinteraktion.csv, topic.csvの\n";
	print "          オブジェクト名を翻訳ファイルのものに変更します\n";
	print "オプション:\n";
	print "    -t|--text   ファイル名  指定したファイルから翻訳テキストを読み込みます\n";
	print "                            (デフォルト: script_text.txt)\n";
	print "    -i|--indir  ディレクトリ名\n";
	print "                            指定したディレクトリからスクリプトを読み込みます\n";
	print "    -o|--outdir ディレクトリ名\n";
	print "                            指定したディレクトリにテキストを変更したファイルを\n";
	print "                            作成します\n";
	print "                            (デフォルト: SCRIPT_NEW)\n";
	print "    -?|--help               使用方法を表示します\n";
}

while (my $var = Encode::decode($os_enc, shift @ARGV)) {
	if ($var =~ /^-/) {
		if ($var eq '-?' or $var eq '--help') {
			usage();
			exit(1);
		} elsif ($var eq '-t' or $var eq '--text') {
			if ($#ARGV < 0) {
				printf("ERROR: オプション%sの後にファイル名がありません\n", $var);
				exit(1);
			}
			$text_file = Encode::decode($os_enc, shift @ARGV);
		} elsif ($var eq '-i' or $var eq '--indir') {
			if ($#ARGV < 0) {
				printf("ERROR: オプション%sの後にディレクトリ名がありません\n", $var);
				exit(1);
			}
			$in_dir = Encode::decode($os_enc, shift @ARGV);
		} elsif ($var eq '-o' or $var eq '--outdir') {
			if ($#ARGV < 0) {
				printf("ERROR: オプション%sの後にディレクトリ名がありません\n", $var);
				exit(1);
			}
			$out_dir = Encode::decode($os_enc, shift @ARGV);
		} else {
			printf("ERROR: 不正なオプションです: %s\n", $var);
			exit(1);
		}
	} else {
		printf("ERROR: 不要なコマンドライン引数です: %s\n", $var);
		exit(1);
	}
}

if ($in_dir ne '') {
	$npc_file                = "$in_dir/$npc_file";
	$inventory_object_file   = "$in_dir/$inventory_object_file";
	$object_interaction_file = "$in_dir/$object_interaction_file";
	$topic_file              = "$in_dir/$topic_file";
}

if ($out_dir ne '') {
	$out_npc_file                = "$out_dir/$out_npc_file";
	$out_inventory_object_file   = "$out_dir/$out_inventory_object_file";
	$out_object_interaction_file = "$out_dir/$out_object_interaction_file";
	$out_topic_file              = "$out_dir/$out_topic_file";
}

my %object_names;
my %ja_text;
my %ja_text2;

sub add_ja_text($$)
{
	my ($text, $ja_text) = @_;
	return if ($text eq '');
	return if ($ja_text eq '');
	return if ($ja_text eq '0');
	return if ($ja_text eq '機械翻訳');
	return if ($ja_text eq '原文');

	if (not exists $ja_text{$text}) {
		$ja_text{$text} = $ja_text;
	}
}

sub add_ja_text2($$)
{
	my ($text, $ja_text) = @_;
	return if ($text eq '');
	return if ($ja_text eq '');
	return if ($ja_text eq '0');
	return if ($ja_text eq '機械翻訳');
	return if ($ja_text eq '原文');

	if (not exists $ja_text2{$text}) {
		$ja_text2{$text} = $ja_text;
	}
}

sub add_object_name($$$$$$)
{
	my ($line, $type, $id, $text, $ja_text, $ja_text2) = @_;
	my $key = "$type\t$id";

	if (exists $object_names{$key}) {
		my ($prev_line, $prev_text, $prev_ja_text, $prev_ja_text2) = split("\t", $object_names{$key}, 4);
		printf("WARNING: 同じ番号の名前が存在します: %d行目, %d行目: %s\n", $prev_line, $line, $id);
	} else {
		$object_names{$key} = join("\t", $line, $text, $ja_text, $ja_text2);
	}

	add_ja_text($text, $ja_text);
	add_ja_text2($text, $ja_text2);
}

sub get_object_name($$$$)
{
	my ($line, $type, $id, $text) = @_;
	my $key = "$type\t$id";

	$text =~ s/[ \t]+$//;

	if (not exists $object_names{$key}) {
		# 翻訳ファイルに存在しない名前
		return '';
	}

	my ($found_line, $found_text, $ja_text, $ja_text2) = split("\t", $object_names{$key}, 4);

	if ($text ne $found_text) {
		printf("WARNING: 翻訳ファイルとスクリプトファイルの原文が違います\n");
		printf("\tスクリプト:%8.1d行目\t%s\n", $line, $text);
		printf("\t翻訳      :%8.1d行目\t%s\n", $found_line, $found_text);
	}

	if ($ja_text eq '') {
		return $ja_text2;
	} elsif ($ja_text eq '原文') {
		return '';
	} elsif ($ja_text eq '機械翻訳') {
		return $ja_text2;
	} elsif ($ja_text ne '0') {
		return $ja_text;
	}

	if (exists $ja_text{$text}) {
		return $ja_text{$text};
	}

	if (exists $ja_text2{$text}) {
		return $ja_text2{$text};
	}

	return '';
}

my $num_npc_replaced = 0;
my $num_inventory_object_replaced = 0;
my $num_object_interaction_replaced = 0;
my $num_topic_replaced = 0;

print "翻訳ファイル読み込み中...\n";
open(my $TEXT, '<:unix:encoding(UTF-8)', $text_file) or die "ERROR: ファイルを開けません: $text_file: " . DEC($!) . "\n";
my $line = 0;
while (<$TEXT>) {
	$line++;

	s/[\r\n]+//;
	next if (m/^#/);
	next if (m/^$/);

	my @work = split("\t", $_);
	if ($#work < 2) {
		printf("ERROR: 翻訳テキストの項目数が不足しています: %d行目\n", $line);
		exit(1);
	}

	my $type = $work[0];
	my $id = $work[1];
	my $text = $work[2];
	my $ja_text = '';
	my $ja_text2 = '';
	if (3 <= $#work) {
		$ja_text = $work[3];
	}
	if (4 <= $#work) {
		$ja_text2 = $work[4];
	}

	next if ($ja_text eq '' and $ja_text2 eq '');

	if ($type eq 'npc' or $type eq 'inventory' or $type eq 'object' or $type eq 'topic') {
		add_object_name($line, $type, $id, $text, $ja_text, $ja_text2);
	} else {
		printf("ERROR: 不正な翻訳テキストです: %d行目\n", $line);
		exit(1);
	}
}
close($TEXT);

if (not -d $out_dir) {
	print "出力ディレクトリ作成\n";
	mkdir($out_dir) or die "ERROR: ディレクトリを作成できません: $out_dir: " . DEC($!) . "\n";
}

print "NPCファイル置換中...\n";
open(my $IN, '<:unix:encoding(UTF-8)', $npc_file) or die "ERROR: ファイルを開けません: $npc_file: " .DEC($!) . "\n";
open(my $OUT, '>:unix:encoding(UTF-8)', $out_npc_file) or die "ERROR: ファイルを開けません: $out_npc_file: " . DEC($!) . "\n";
$line = 0;
while (<$IN>) {
	$line++;
	if (m#^([0-9]+)(;[0-9]+;[0-9]+;)([^;]+)(;[^;]+;[0-9]+(?:\.[0-9]+)?;[0-9]+(?:\.[0-9]+)?;[0-9]+;[0-9]+)([ \t]*[\r\n]+)$#) {
		my ($id, $unknown1, $text, $unknown2, $crlf) = ($1, $2, $3, $4, $5);
		my $ja_text = get_object_name($line, 'npc', $id, $text);
		if ($ja_text eq '') {
			print $OUT $_;
		} else {
			if ($ja_text =~ m/^<#>/) {
				$ja_text =~ s/^<#>/$id/;
			}
			print $OUT $id, $unknown1, $ja_text, $unknown2, $crlf;
			$num_npc_replaced++;
		}
	} else {
		print $OUT $_;
	}
}
close($OUT);
close($IN);

print "持ち物ファイル置換中...\n";
open(my $IN, '<:unix:encoding(UTF-8)', $inventory_object_file) or die "ERROR: ファイルを開けません: $inventory_object_file: " .DEC($!) . "\n";
open(my $OUT, '>:unix:encoding(UTF-8)', $out_inventory_object_file) or die "ERROR: ファイルを開けません: $out_inventory_object_file: " . DEC($!) . "\n";
$line = 0;
while (<$IN>) {
	$line++;
	if (m#^([0-9]+)(;[0-9]+;)([^;]+)(;(?:items|topics)/[^;/]+;[0-9]+;[^;]+;[0-9]+;[0-9]+;[0-9]+)([ \t]*[\r\n]+)$#) {
		my ($id, $unknown1, $text, $unknown2, $crlf) = ($1, $2, $3, $4, $5);
		my $ja_text = get_object_name($line, 'inventory', $id, $text);
		if ($ja_text eq '') {
			print $OUT $_;
		} else {
			if ($ja_text =~ m/^<#>/) {
				$ja_text =~ s/^<#>/$id/;
			}
			print $OUT $id, $unknown1, $ja_text, $unknown2, $crlf;
			$num_inventory_object_replaced++;
		}
	} else {
		print $OUT $_;
	}
}
close($OUT);
close($IN);

print "オブジェクトファイル置換中...\n";
open(my $IN, '<:unix:encoding(UTF-8)', $object_interaction_file) or die "ERROR: ファイルを開けません: $object_interaction_file: " .DEC($!) . "\n";
open(my $OUT, '>:unix:encoding(UTF-8)', $out_object_interaction_file) or die "ERROR: ファイルを開けません: $out_object_interaction_file: " . DEC($!) . "\n";
$line = 0;
while (<$IN>) {
	$line++;
	if (m#^([0-9]+)(;[0-9]+;)([^;]+)(;[0-9]+;[0-9]+;[^;]+;[^;]+;[0-9]+;[0-9]+;[0-9]+;[0-9]+)([ \t]*[\r\n]+)$#) {
		my ($id, $unknown1, $text, $unknown2, $crlf) = ($1, $2, $3, $4, $5);
		my $ja_text = get_object_name($line, 'object', $id, $text);
		if ($ja_text eq '') {
			print $OUT $_;
		} else {
			if ($ja_text =~ m/^<#>/) {
				$ja_text =~ s/^<#>/$id/;
			}
			print $OUT $id, $unknown1, $ja_text, $unknown2, $crlf;
			$num_object_interaction_replaced++;
		}
	} else {
		print $OUT $_;
	}
}
close($OUT);
close($IN);

print "話題ファイル置換中...\n";
open(my $IN, '<:unix:encoding(UTF-8)', $topic_file) or die "ERROR: ファイルを開けません: $topic_file: " .DEC($!) . "\n";
open(my $OUT, '>:unix:encoding(UTF-8)', $out_topic_file) or die "ERROR: ファイルを開けません: $out_topic_file: " . DEC($!) . "\n";
$line = 0;
while (<$IN>) {
	$line++;
	if (m#^([0-9]+)(;[0-9]+;)([^;]+)(;topics/[^;/]+;[0-9]+;[0-9]+;[0-9]+)([ \t]*[\r\n]+)$#) {
		my ($id, $unknown1, $text, $unknown2, $crlf) = ($1, $2, $3, $4, $5);
		my $ja_text = get_object_name($line, 'topic', $id, $text);
		if ($ja_text eq '') {
			print $OUT $_;
		} else {
			if ($ja_text =~ m/^<#>/) {
				$ja_text =~ s/^<#>/$id/;
			}
			print $OUT $id, $unknown1, $ja_text, $unknown2, $crlf;
			$num_topic_replaced++;
		}
	} else {
		print $OUT $_;
	}
}
close($OUT);
close($IN);

print "完了\n\n";

printf("NPC         : %d件置換\n", $num_npc_replaced);
printf("持ち物      : %d件置換\n", $num_inventory_object_replaced);
printf("オブジェクト: %d件置換\n", $num_object_interaction_replaced);
printf("話題        : %d件置換\n", $num_topic_replaced);

